from abc import ABC, abstractmethod
from pydantic import BaseModel
from typing import get_args, get_origin

from llama_prompter.types import Concat


class BNF(ABC):
    BASE_TYPES: dict[type, tuple[str, str]]

    @classmethod
    def type_name(cls, type_: type) -> str:
        if type_ in cls.BASE_TYPES:
            return cls.BASE_TYPES[type_][0]
        origin = get_origin(type_)
        if origin:
            return f"{origin.__name__.lower()}" + "".join([cls.type_name(a).title() for a in get_args(type_)])
        return type_.__name__.lower()

    @classmethod
    @abstractmethod
    def definition(cls, type_: type) -> str:
        pass

    @staticmethod
    @abstractmethod
    def rule(name: str, definition: str) -> str:
        pass


class EBNF(BNF):
    BASE_TYPES = {
        bool: ("boolean", r'"true" / "false"'),
        float: ("float", r"~'-?[0-9]*(.[0-9]*)?'"),
        int: ("integer", r"~'-?[0-9]+'"),
        str: ("string", r"~'\"[0-9 a-z]*\"'i"),
    }

    @classmethod
    def definition(cls, type_: type) -> str:
        if type_ in cls.BASE_TYPES:
            return cls.BASE_TYPES[type_][1]
        elif get_origin(type_) == Concat:
            return ' " " '.join([cls.type_name(a) for a in get_args(get_args(type_)[0])])
        elif get_origin(type_) == list:
            arg = cls.type_name(get_args(type_)[0])
            return f'"[" ({arg} ("," {arg})*)? "]"'
        elif get_origin(type_) == tuple:
            return '"[" ' + ' "," '.join([cls.type_name(a) for a in get_args(type_)]) + ' "]"'
        elif get_origin(type_) == dict:
            k_type, v_type = get_args(type_)
            kv = f'{cls.type_name(k_type)} ":" {cls.type_name(v_type)}'
            return f'"{{" ({kv} ("," {kv})*)? "}}"'
        elif issubclass(type_, BaseModel):
            return '"{" ' + ' "," '.join([f'"\\"{n}\\":" {cls.type_name(f.annotation)}' for n, f in type_.model_fields.items() if f.annotation]) + ' "}"'
        assert False, f"Unable to return bnf definition for type {type_.__name__}"

    @staticmethod
    def rule(name: str, definition: str) -> str:
        return f"{name} = {definition}"


class GBNF(BNF):
    BASE_TYPES = {
        bool: ("boolean", '("true" | "false")'),
        float: ("float", '("-"? ([0-9]*) ("." [0-9]+))'),
        int: ("integer", '("-"? ([0-9]+))'),
        str: ("string", '"\\"" ([^"\\\\] | "\\\\" (["\\\\/bfnrt] | "u" [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F]))* "\\""'),
    }

    @classmethod
    def definition(cls, type_: type) -> str:
        if type_ in cls.BASE_TYPES:
            return cls.BASE_TYPES[type_][1]
        else:
            return EBNF.definition(type_)

    @staticmethod
    def rule(name: str, definition: str) -> str:
        return f"{name} ::= {definition}"



import json
import re

from pydantic import BaseModel
from typing import Any, Callable, Optional, TypeVar, Union, get_args, get_origin

from llama_prompter.bnf import BNF, EBNF
from llama_prompter.types import Concat


def parse_type(bnf: type[BNF], rules: dict[str, str], type_: type) -> str:
    type_name = bnf.type_name(type_)
    if type_name in rules:
        return type_name
    rules[type_name] = bnf.definition(type_)
    if get_origin(type_) == Concat:
        for x in get_args(get_args(type_)[0]):
            parse_type(bnf, rules, x)
    elif get_origin(type_) in [dict, list, tuple]:
        for x in get_args(type_):
            parse_type(bnf, rules, x)
    elif issubclass(type_, BaseModel):
        for _, f in type_.model_fields.items():
            if f.annotation:
                parse_type(bnf, rules, f.annotation)
    return type_name


def type_to_grammar(type_: type, bnf: type[BNF] = EBNF) -> str:
    rules: dict[str, str] = {}
    rules["root"] = parse_type(bnf, rules, type_)
    return "\n".join([bnf.rule(n, r) for n, r in rules.items()])


BASE_TYPES_RE: dict[type, tuple[re.Pattern[str], Callable[[str], Any]]] = {
    bool: (re.compile(r"(true|false)"), lambda x: x == "true"),
    float: (re.compile(r"(\-?[0-9]*\.[0-9]+)"), float),
    int: (re.compile(r"(\-?[0-9]+)"), int),
    str: (re.compile(r'"((?:\\.|[^"\\])*)"'), str),
}

LIST_TYPE_RE = re.compile(r"\[\]")

T = TypeVar("T")


def decode_type(payload: str, type_: type[T]) -> tuple[Optional[T], int]:
    if type_ in BASE_TYPES_RE:
        regex, cast_func = BASE_TYPES_RE[type_]
        m = regex.match(payload)
        if not m:
            return None, 0
        return cast_func(m.group(1)), len(m.group(0))
    elif get_origin(type_) == Concat:
        # queue.extend(get_args(get_args(type_)[0]))
        pass
    elif get_origin(type_) == list:
        start = 1  # skip first "["
        result = []
        arg_type = get_args(type_)[0]
        while payload[start - 1] != "]":
            if start >= len(payload):
                return None, 0
            arg, i = decode_type(payload[start:], arg_type)
            if arg is None:
                return None, 0
            result.append(arg)
            start += i + 1  # skip comma ","
        return list(result), start  # type: ignore
    elif get_origin(type_) == tuple:
        start = 1  # skip first "["
        result = []
        for arg_type in get_args(type_):
            if start >= len(payload):
                return None, 0
            arg, i = decode_type(payload[start:], arg_type)
            if arg is None:
                return None, 0
            result.append(arg)
            start += i + 1  # skip comma ","
        return tuple(result), start  # type: ignore
    elif issubclass(type_, BaseModel):
        d, i = json.JSONDecoder().raw_decode(payload)
        return type_(**d), i  # type: ignore
    return None, 0


def deserialize(payload: str, type_: type[T]) -> T:
    obj, i = decode_type(payload, type_)
    if obj is None:
        raise Exception("Could not decode payload")
    return obj



import re
import inspect

from typing import Any, Optional, Union
from llama_cpp.llama_grammar import LlamaGrammar

from llama_prompter.parser import parse_type, decode_type
from llama_prompter.bnf import BNF, GBNF

TEMPLATE_VAR_RE = re.compile(r"(?<!{){([a-zA-Z_]+):([^{}]*)}(?!})")


class Variable:
    def __init__(self, name: str, type_: type) -> None:
        self._name = name
        self._type = type_


def _type_to_grammar(inputs: list[Union[str, Variable]]) -> str:
    rules: dict[str, str] = {}
    rules["root"] = " ".join([parse_type(GBNF, rules, x._type) if isinstance(x, Variable) else f'"{x}"' for x in inputs])
    return "\n".join([GBNF.rule(n, r) for n, r in rules.items()])


def get_caller_symbols() -> tuple[dict[str, str], dict[str, str]]:
    frame = inspect.currentframe()
    if frame is None or frame.f_back is None or frame.f_back.f_back is None:
        raise Exception("Could not retrieve caller's symbols")
    return frame.f_back.f_back.f_globals, frame.f_back.f_back.f_locals


class Prompt:
    def __init__(self, template: str) -> None:
        self._prompt = ""
        self._sequences: list[Union[str, Variable]] = []
        caller_g, caller_l = get_caller_symbols()
        matches = [x for x in TEMPLATE_VAR_RE.finditer(template)]
        for i, m in enumerate(matches):
            if i == 0:
                self._prompt = template[0 : m.start(0)]
            else:
                self._sequences.append(template[matches[i - 1].end(0) : m.start(0)])
            type_ = eval(m[2], caller_g, caller_l)
            assert isinstance(type_, type)
            var = Variable(m[1], type_)
            self._sequences.append(var)
        if matches and matches[-1].end(0) < len(template):
            self._sequences.append(template[matches[-1].end(0) :])
        self._grammar = LlamaGrammar.from_string(_type_to_grammar(self._sequences), verbose=False) if self._sequences else None

    @property
    def prompt(self) -> str:
        return self._prompt

    @property
    def grammar(self) -> Optional[LlamaGrammar]:
        return self._grammar

    def decode_response(self, response: str) -> dict[str, Any]:
        variables: dict[str, Any] = {}
        start = 0
        for i in self._sequences:
            if isinstance(i, str):
                if not response[start:].startswith(i):
                    raise Exception("Could not decode")
                start += len(i)
            elif isinstance(i, Variable):
                obj, idx = decode_type(response[start:], i._type)
                if obj is None:
                    raise Exception("Could not decode")
                variables[i._name] = obj
                start += idx
        return variables
